/* Freqt.java
   Copyright 2002, 
   Tatsuya Asai, Kenji Abe, Shinji Kawasoe, Hiroshi Sakamoto,
   Hiroki Arimura and Setsuo Arikawa,
   all rights reserved.

   This file contains an implementation of the algorithm "FREQT".

   This software may be freely used for academic purpose only. Please
   contact us for profit use. However, when distributed, the original
   source must be clearly stated, and, when the source code is
   distributed, the copiright notice must be retained and any
   alterations in the code must be clearly marked. No warranty is
   given regarding the quality of this software.
*/

import java.io.*;
import java.lang.*;
import java.util.*;
import org.w3c.dom.*;

/**
 * FREQT 
 */

public class Freqt {

    //Constants
    static final int UNDEF = -1;
    static final boolean DEBUG = false;
    static boolean DEBUG_TRACE = false;
    static final boolean DEBUG_NONE = false; //To be always false

    //Global variables
    static int gStage    	= 1;
    static int gTreeSize 	= 0;    // Datatree Size
    static int gMinHit        	= 0;    // Minimum support
    static double gMinSup 	= 0.0;	// Minimum support (%)
    //static long gTotalCpuTime 	= 0;	// Total Runtime (sec)
    static CpuTimer gTotalTimer = null;
    static String gOutputFile   = null;
    static Node gDomRoot        = null;
    static int gTotalNumFreq   	= 0;    // Total Number of Patterns found

    static HashSet ForbiddenNodeSet   = new HashSet();   // for Node-Skip
    static HashMap ForbiddenEdgeSet   = new HashMap();   // for Edge-Skip
    static HashMap ctPathSet      = new HashMap();   // for Concept Tree

    // Options and default settings
    static boolean ENABLE_CLOSED_PATTERNS     = false;
    static boolean ENABLE_ATTR_NODE           = false;
    static boolean ENABLE_VALUE_NODE           = false;
    //Not implemented
    static boolean ENABLE_CONCEPT_NODE        = false;
    //SDM2002
    static boolean ENABLE_DUPLICATE_DETECTION = true;
    static boolean ENABLE_PRINT_PATTERNS      = true;
    static boolean ENABLE_XML_FORMAT          = false;
    static boolean ENABLE_PRUN_1NODE = false;
    static boolean ENABLE_PRUN_2NODE = false;

    // Debugging and Reporting
    static boolean DEBUG_PRINT_ARGS  = true;
    static boolean DEBUG_PRINT_STAT  = true;
    static boolean DEBUG_PRINT_DOMTREE  = false; 

    final static String VERSION_STRING =
    "@ Freqt: A fast frequent tree miner for XML/HTML documents\n"+
    "@ - Hiroki Arimura, Kenji Abe, Tatsuya Asai, Mar. 2002. "+
    "All rights reserved.\n"+
    "@ - Contact: {arim,asai}@i.kyushu-u.ac.jp, Department of Informatics"; 
    //+"$Id:";

    final static String USAGE_STRING =
    "USAGE: java Freqt infile.xml minsup [OPTIONS]\n"+
    " infile: the input HTML/XML file\n"+
    " minsup: the minimum-support (%)\n"+
    " <OPTIONS>\n"+
    //"  -a | -attr     \tEnable Attribute Expand (possible bug?)\n"+
    "  -o | -out <outfile.xml> \t Output file (e.g. output.xml)\n"+
    "  -m | -maxsize <size>    \t Maximum size of patterns\n"+
    "  -t | -trace   \tSet trace on (for debug)\n"+
    //"  -noprint       \tDisable printing frequent patterns\n"+
    //"  -c <concept>   \tSet Concept Tree with <CONCEPT> (not implemented yet)\n"+
    " Default setting: -dd -noattr.\n"; 

    static void read_options(String[] args, int argc) {
		boolean setDD, setAttr, setConcept, setLimit;
		setDD = setAttr = setConcept = setLimit = false;

		while (argc < args.length){
			// Setting for Duplicate-Detection
			if (args[argc].equals("-dd")){
				argc++;
				ENABLE_DUPLICATE_DETECTION = true;
				setDD = true;
			}
			else if (args[argc].equals("-nodd")){
				argc++;
				ENABLE_DUPLICATE_DETECTION = false;
				setDD = true;
			}
			// Setting for Attribute
			else if (args[argc].equals("-a") || args[argc].equals("-attr")){
				argc++;
				ENABLE_ATTR_NODE   = true;
				DomWrapper.setEnableAttachAttr(true);
				setAttr = true;
			}
			else if (args[argc].equals("-v") || args[argc].equals("-value")){
				argc++;
				ENABLE_ATTR_NODE = true;
				DomWrapper.setEnableAttachAttr(true);
				setAttr = true;
			}
			else if (args[argc].equals("-noattr")){
				argc++;
				ENABLE_ATTR_NODE = false;
				DomWrapper.setEnableAttachAttr(false);
				setAttr = true;
			}
			else if (args[argc].equals("-p")){
				argc++;
				DEBUG_PRINT_DOMTREE = true;
			}
			else if (args[argc].equals("-o") || args[argc].equals("-out")){
				argc++;
				gOutputFile = args[argc++];
			}
			// Setting for Upper Concept
			else if (args[argc].equals("-c")){
				argc++;
				ENABLE_CONCEPT_NODE = true;
				gConceptFile = new File(args[argc++]);
				setConcept = true;
			}
			// Setting for stage limit	
			else if (args[argc].equals("-m") || args[argc].equals("-maxsize")){
				argc++;
				gMaxStage = Integer.parseInt(args[argc++]);
				setLimit = true;
			}
			else if (args[argc].equals("-t") || args[argc].equals("-trace")){
				argc++;
				DEBUG_TRACE = true;
			}
			// Other case
			// Setting for stage limit	
			else if (args[argc].equals("-noprint")){
				//if (setLimit == true)
				//	throw new IllegalArgumentException();
				argc++;
				ENABLE_PRINT_PATTERNS = false;
			}
			else if (args[argc].equals("-xml")){
				argc++;
				ENABLE_XML_FORMAT = true;
			}
			// Other case
			else
				throw new IllegalArgumentException();
		}
    }

    static File   gConceptFile 	= null;
    static int    gMaxStage 	= UNDEF;
    static File   gPageFile 	= null;
    static PrintStream gOutStream = null;
    static double gMinSupInPercent = 0;

    static PrintStream openOutStream(String outputFile) {
		PrintStream outStream = null;
		try {
			if (gOutputFile == null) 
				outStream = System.out;
			else            
				outStream =
					new PrintStream(
									new BufferedOutputStream(
															 new FileOutputStream(outputFile))
									);
		}
		catch(IOException e){
			Util.errorExit("main: cannot open outputFile: "+outputFile);
		}
		return outStream;
    }

    static String getArgsString() {
		return "@infile\t"+gPageFile.getName()+
			"\n@size \t"+gPageFile.length()+" (byte)"+ 
			"\t"+gPageFile.length()/1000+" (MB)"+ 
			"\n@args minsup\t" + gMinSupInPercent+" (%)"; 
    }

    static void printArgs() {
		if (DEBUG_PRINT_ARGS) {
			Util.task("Print arguments...");
			System.out.println(getArgsString());
		}
    }

    static String getDomStatString() {
		return
			"@stat size of datatree\t"+gTreeSize+" (nodes)\n"+
			"@stat minsup \t"+gMinSup+" (%)\n"+
			"@stat minhit \t"+gMinHit+" (nodes)";
    }

    static void printDomStat() {
        if (DEBUG_PRINT_STAT) {
			System.out.println(getDomStatString());
		}
    }

    public static void main(String[] args){
        try{
			int argc = 0;

            System.out.println(VERSION_STRING);
			//Reading arguments and options
			if (args.length < 2)
				throw new IllegalArgumentException();
			if (args.length < 2)
				throw new IllegalArgumentException();
			gPageFile = new File(args[argc++]);
			gMinSupInPercent = Double.parseDouble(args[argc++]);
			gMinSup = gMinSupInPercent * 0.01;

			read_options(args, argc);
			printArgs();

			// Program Main Routine

			Util.task("Building DOM-tree...");
			gDomRoot = buildDomTree(gPageFile, gConceptFile);
			long gTreeSize = DomWrapper.getTreeSize();
			gMinHit = (int) Math.ceil(gTreeSize*gMinSup);
			printDomStat();

			if (DEBUG_PRINT_DOMTREE) {
				PrintDomTree.printTree(gDomRoot, DomWrapper.getDocument());
				System.exit(0);
			}

			Util.task("Mining Patterns...");
			gTotalTimer = new CpuTimer(CpuTimer.OFF);
			gOutStream = openOutStream(gOutputFile);

			if (gOutputFile != null)
				gOutStream.println(
								   getXMLCommentString("\n"+VERSION_STRING+"\n")+"\n"+
								   getXMLCommentString("\n"+getArgsString()+"\n")+"\n"+
								   getXMLCommentString("\n"+getDomStatString()+"\n")+"\n"
								   );

			Search.execute(gDomRoot);

			gTotalTimer.print("@total time \t");
			System.out.println("@total #frequents\t"+gTotalNumFreq);
			gOutStream.close();
        }
		catch(IllegalArgumentException e){
			//e.printStackTrace();
            System.err.println("\nToo few/many arguments or wrong options.\n"+
							   USAGE_STRING);
            System.exit(1);
        }
        catch(Exception e){
			e.printStackTrace();
            System.err.println(e);
            System.exit(1);
        }
    }

	
    //
    // Create DOM Tree
    //
    static
    Node buildDomTree(File pagefile, File conceptfile) throws Exception{

		CpuTimer parseTimer = new CpuTimer(CpuTimer.ON);

		DomWrapper.initialize();
		Node root = DomWrapper.createDom(pagefile);
		Document doc = DomWrapper.getDocument();

		parseTimer.stop();
		parseTimer.print("@time (parse) \t");

		Util.task("Attaching attributes...");

		//if (ENABLE_ATTR_NODE)
		//    DomWrapper.createAttrTree(root, doc);
	
		if (ENABLE_CONCEPT_NODE)
			setCTPath(conceptfile);

		parseTimer.stop();
		parseTimer.print("@time (attrs) \t");

		return root;
    }

	    
    /**************************************************
     * Scan the Dom Tree
     **************************************************/

    static void scanDomTree(Node domNode, HashMap patHashMap){
        Pattern newpat = null;
        String label = domNode.getNodeName();
		gTreeSize++;

        if ((newpat = (Pattern) patHashMap.get(label)) == null) {
            //newpat = pattern.getRMExpansion(p, domNode);
			Snode theNewNode = new Snode(domNode); //The node to be attached
			theNewNode.setNodeValue(null);
            newpat = new Pattern(theNewNode); //Create a copy of domnode
            patHashMap.put(label, newpat);
        }
		newpat.addRmOcc(domNode);

        Node childDomNode = domNode.getFirstChild();
        while (childDomNode != null){
            scanDomTree(childDomNode, patHashMap);
            childDomNode = childDomNode.getNextSibling();
        }
    }

    /**************************************************
     * Search Main routine
     * => See the separate file Search.java 
     **************************************************/

    /**************************************************
     * doRightmostExpansion
     **************************************************/

    /////////////////////////////////////////////////////////////////////
    // doRightmostExpansion: Given a pattern and an occurrence oldRmoNode
    // of its rightmost leaf, scan the children of the branchNode
    // starting from startNode.
    // Rightmost Expansion
    //         o
    //        /
    //       o:branchNode
    //      / \
    //     /   \
    //    /     o:startNode
    //   o:_headNode
    //  /
    // /
    //o:oldRmoNode
    /////////////////////////////////////////////////////////////////////

    static void doRightmostExpansion(Pattern pattern, PatternList newPatList, int p){
		//Note: The number p is the depth of the branching node.

		HashMap patHashMap = new HashMap();
		Snode branch = pattern.fetchPthParent(p);
		if (branch.getNodeType() == Node.ELEMENT_NODE) {
			if (DEBUG_TRACE) System.out.println("\nDO EXPANSION: ELEMENT NODE: "+
												branch.getNodeName());
			expandElementNodes(pattern, p, patHashMap);
		}
		else if (ENABLE_ATTR_NODE) {
			if (DEBUG_TRACE) System.out.println("\nDO EXPANSION: NON-ELEMENT NODE: "+
												branch.getNodeName());
			expandValueNodes(pattern, p, patHashMap);
		}
		else
			;/* Do nothing */
		newPatList.addAll(patHashMap.values());
    }

    private static void expandValueNodes(Pattern pattern, int p, HashMap patHashMap) {
		if (p > 0)
			Util.errorExit("expandValueNodes: p must be zero!: p="+p);
		if (DEBUG_TRACE && gStage > 1) debugDRE1(pattern, p);

		if (pattern == null)
			Util.errorExit("expandValueNodes: null pattern!");
		if (pattern.getRmLeaf().getNodeValue() != null) {
			//The value has been already expanded
			return;
		}

		NodeList rmOccList = pattern.getRmoList();
		for (int i = 0; i < rmOccList.size(); i++){
			Node   rmoNode  = (Node)rmOccList.get(i);
			if (DEBUG_TRACE && gStage > 1) 
				System.out.println("SCAN RMOs: ["+i+"] <"+
								   rmoNode.getNodeName()+">");
			String value = UserData.getVirtualValue(rmoNode);
			if (value != null) {
				Pattern theExpansion =
					lookupAnExpansionOnVALUE(pattern, value, rmoNode, patHashMap);
				theExpansion.addRmOcc(rmoNode);
			}
		}//for
    }

    static Pattern lookupAnExpansionOnVALUE(Pattern oldpat, String value, Node domNode, HashMap patHashMap) {
		Pattern newpat = null;

		if ((newpat = (Pattern) patHashMap.get(value)) == null) {
			if (DEBUG_TRACE)
				System.out.println("  VALUE "+oldpat.getRootNode()+
								   " with "+"\""+value+"\"");

			newpat = new Pattern();          //Create a copy of oldpat
			newpat.copySkeltonOnly(oldpat);
			Snode theNode = newpat.getRmLeaf(); //The rightmost leaf of newpat
			theNode.setNodeValue(value);
			patHashMap.put(value, newpat);

			if (DEBUG_TRACE) {
				System.out.println("  VALUE ATTACHED => "+newpat);
			}
		}
		else {
			if (DEBUG_TRACE) {
				System.out.println("  INCREMENTED => rmoOcc="+(newpat.getRmOcc()+1));
			}
		}
		return newpat;
    }

    private static void expandElementNodes(Pattern pattern, int p, HashMap patHashMap) {
		if (DEBUG_TRACE && gStage > 1) debugDRE1(pattern, p);

		Node last = null; //For DD: Duplicate detection
		NodeList rmOccList = pattern.getRmoList();
		for (int i = 0; i < rmOccList.size(); i++){

			Node   oldRmoNode  = (Node)rmOccList.get(i);
			Node _headNode, branchNode, startNode;

			if (DEBUG_TRACE && gStage > 1) {
				System.out.println("SCAN RMOs: ["+i+"] <"+oldRmoNode.getNodeName()+">");
			}

			//AtParentNode (p > 0)
			if (p == 0) {
				startNode = _headNode = oldRmoNode.getFirstChild(); 
				//Note: startNode and _headNode can be null but OK. 
				branchNode = oldRmoNode; //Self
				//Note: The _headNode may be null when p=0;
			}
			else {
				_headNode= getAncestorNode(p-1, oldRmoNode); //Aux
				branchNode = _headNode.getParentNode(); //Branching point
				startNode  = _headNode.getNextSibling();
				//Note: The _headNode and branchNode must be nonnull when p>0;
			}

			if (DEBUG_TRACE && gStage > 1 ) 
				debugDRE3_BRANCH(_headNode, branchNode, startNode, p);

			if (branchNode == last) //For DD: Duplicate detection
				continue; //for - skip the rest
			//Process the new RMO node by scanning
			ScanSiblingsOfElementNode(pattern, startNode, branchNode, p, 
									  patHashMap);
			last = branchNode; //For DD: Duplicate detection
		}//for
    }

    static void ScanSiblingsOfElementNode(Pattern pattern, Node startNode, Node branchNode, int p, HashMap patHashMap) {
		//Process the new RMO node by scanning

		Node newRmoNode = startNode;
		while (newRmoNode != null){
			if (DEBUG_TRACE && gStage > 1) {
				System.out.println("  SCAN CHILDREN: <"+newRmoNode.getNodeName()+">");
			}

			String label = newRmoNode.getNodeName();
			if (branchNode.getNodeType() == Node.ELEMENT_NODE) {

				if (ENABLE_PRUN_1NODE &&
					!isSafeNode(label))
					continue;
				if (ENABLE_PRUN_2NODE &&
					!isSafeEdge(branchNode.getNodeName(), label))
					continue;

				Pattern theExpansion =
					lookupAnExpansionOnELEMENT(pattern, p, newRmoNode, patHashMap);
				//With DD: No need to check the uniqueness of RMO
				theExpansion.addRmOcc(newRmoNode);
			}
			newRmoNode = newRmoNode.getNextSibling();

		}//while

    }

    static Pattern lookupAnExpansionOnELEMENT(Pattern oldpat, int p, Node domNode, HashMap patHashMap) {
		Pattern newpat = null;
		String label = domNode.getNodeName();

		if ((newpat = (Pattern) patHashMap.get(label)) == null) {
			if (DEBUG_TRACE)
				System.out.println("  EXPAND "+oldpat.getRootNode()+
								   " with "+
								   " <"+p+", \""+domNode.getNodeName()+"\">");
			//newpat = pattern.getRMExpansion(p, domNode);
			newpat = new Pattern();          //Create a copy of oldpat
			newpat.copySkeltonOnly(oldpat);
			Snode theNewNode = new Snode(domNode); //The node to be attached
			theNewNode.setNodeValue(null);
			newpat.expand(p, theNewNode); //Doing rightmost expansion
			patHashMap.put(label, newpat);

			if (DEBUG_TRACE) {
				System.out.println("  CHILD CREATED => "+newpat);
			}
		}
		else {
			if (DEBUG_TRACE) {
				System.out.println("  INCREMENTED => rmoOcc="+(newpat.getRmOcc()+1));
			}
		}
		return newpat;
    }

    static void debugDRE1(Pattern pattern, int p) {
		String label = pattern.getRootNode().getNodeName();
		System.out.println("\nPARENT: "+pattern+
						   "\n  root="+pattern.getRootNode()+
						   "\n  rml="+pattern.getRmLeaf()+
						   "\nDEPTH: "+p+
						   "\n");
    }

    static void debugDRE3_BRANCH(Node _headNode, Node branchNode, Node startNode, int p) {
		String headname, startname;
		if (_headNode == null)
			headname = null;
		else
			headname = _headNode.getNodeName();
		if (startNode == null)
			startname = null;
		else
			startname = _headNode.getNodeName();
		if (branchNode == null)
			Util.errorExit("doRightmostExpansion: branchNode==null!");
		System.out.println("  BRANCH COMPUTED: "+
						   "p="+p+
						   ", node="+branchNode.getNodeName()+
						   ", #child="+getNumChildren(branchNode)+
						   ", _head="+headname+
						   ", start="+startname+
						   "");
    }

    static void doRightmostExpansionAtBottom(Pattern pattern, PatternList newPatList, int p){
		//Note: The number p is the depth of the branching node.

		//DD is not needed for the bottom position
		HashMap patHashMap = new HashMap();

		NodeList rmOccList = pattern.getRmoList();
		int pmax = pattern.getRmbLength();
	
		for (int i = 0; i < rmOccList.size(); i++){
			Node   oldRmoNode  = (Node)rmOccList.get(i);

			//At Bottom Position
			Node branchNode = oldRmoNode; //Branching point
			if (branchNode.getNodeType() == Node.ELEMENT_NODE) {
				Node startNode  = oldRmoNode.getFirstChild();
				ScanSiblingsOfElementNode(pattern, startNode, branchNode, 0, patHashMap);
				if (ENABLE_ATTR_NODE) {
					//scanSiblingsAttrNode(pattern, branchNode, patHashMap);
					;
				}
			}
		}//for
		newPatList.addAll(patHashMap.values());
    }

	
    /** Pruning
     */
    static PatternList pruneByRMO(PatternList plist){

		PatternList newPlist = new PatternList();
		for (int i = 0; i < plist.size(); i++){
			Pattern pattern = (Pattern)plist.get(i);
			if (pattern.getRmOcc() >= gMinHit)
				newPlist.add(pattern);
			else
				/* Do not add */;
		}

		return newPlist;
    }

    static PatternList pruneFirstStage(PatternList plist){

		PatternList newPlist = new PatternList();

        for (int i = 0; i < plist.size(); i++){
            Pattern pattern = (Pattern)plist.get(i);

            if (pattern.getRootOcc() >= gMinHit)
				newPlist.add(pattern);

			else{
				if (ENABLE_PRUN_1NODE){
					Snode node = pattern.getRmLeaf();
					String nodename = node.getNodeName();
					addForbiddenNodeSet(nodename);
				}
			}
        }

		return newPlist;
    }

    static PatternList pruneMainStage(PatternList plist){

		PatternList newPlist = new PatternList();
        for (int i = 0; i < plist.size(); i++){
			Pattern pattern = (Pattern)plist.get(i);

            if (pattern.getRootOcc() >= gMinHit){
				newPlist.add(pattern);
			}
			else
				/* Discard the pattern */;
		}
		return newPlist;
    }

		
    static void printFreqPatterns(PatternList plist){

		gOutStream.println(getXMLCommentString("=== Stage " + gStage + " ==="));

		if (ENABLE_PRINT_PATTERNS) {
			for (int i = 0; i < plist.size(); i++){
				Pattern pattern = (Pattern)plist.get(i);
				Snode patnode = pattern.getRmLeaf();
	    
				gOutStream.println("<freqpattern rank="+xmlAValInt(i)+
								   " stage="+xmlAValInt(gStage)+
								   " />");
				if (ENABLE_XML_FORMAT) {
					gOutStream.print(patnode.printTagStyle());
				}
				else {
					gOutStream.println(pattern);
				}
			}
			if (gOutputFile != null) {
				gOutStream.println(getXMLCommentString(
													   "@stage "+gStage+": "+plist.size()+"patterns found"));
				gOutStream.println();
			}
		}
		System.out.println("@stage "+gStage+": "+
						   plist.size()+" patterns found");
    }

    static void printFreqPatterns(PatternList plist, int stage){

		gOutStream.println(getXMLCommentString("=== Stage " + stage + " ==="));

		if (ENABLE_PRINT_PATTERNS) {
			for (int i = 0; i < plist.size(); i++){
				Pattern pattern = (Pattern)plist.get(i);
				Snode patnode = pattern.getRmLeaf();
	    
				gOutStream.println("<freqpattern rank="+xmlAValInt(i)+
								   " stage="+xmlAValInt(stage)+
								   " />");
				if (ENABLE_XML_FORMAT) {
					gOutStream.print(patnode.printTagStyle());
				}
				else {
					gOutStream.println(pattern);
				}
			}
			if (gOutputFile != null) {
				gOutStream.println(getXMLCommentString(
													   "@stage "+stage+": "+plist.size()+"patterns found"));
				gOutStream.println();
			}
		}
		System.out.println("#stage "+stage+": "+
						   plist.size()+" parents found");
    }

    /**************************************************
     * Mining: Computing Root Occurrences
     **************************************************/

    /** Count the root occurrences of the candidate patterns of size one */
    static void computeRootOccFirstStage(PatternList plist){
		for (int i = 0; i < plist.size(); i++){
			Pattern pattern = (Pattern)plist.get(i);
			pattern.setRootOcc(pattern.getRmOcc());
		}
    }

    /** Count the root occurrences of the candidate patterns */
    static void computeRootOcc(PatternList plist){
		for (int i = 0; i < plist.size(); i++){
			Pattern pattern = (Pattern)plist.get(i);
			int rootocc = countRootOcc(pattern);
			pattern.setRootOcc(rootocc);
		}
    }

    /** Computes the distict occurrence of the p-th parent of the RML */
    static int countPthParentOcc(Pattern pattern, int p){
		//int branchDepth = pattern.getRmbLength();
		int branchDepth = p;

		NodeList rmoList = pattern.getRmoList();
		Node lastocc = null;
		int rootCount = 0;
		for (int i = 0; i < rmoList.size(); i++){
			Node rmo = (Node)rmoList.get(i);
			Node target = getAncestorNode(branchDepth, rmo);

			if (target != lastocc){
				rootCount++;
				lastocc = target;
			}
		}
   
		return rootCount;
    }

    static int countRootOcc(Pattern pattern){
		return countPthParentOcc(pattern, pattern.getRmbLength());
    }

    static int org_countRootOcc(Pattern pattern){

		int rmbLength = pattern.getRmbLength();
		NodeList rmoList = pattern.getRmoList();
		// Variables for Duplicate-Detection
		Node lastocc = null;
		int rootCount = 0;
		// Variable for Non Duplicate-Detection
		NodeList rootOccList = new NodeList();

		for (int i = 0; i < rmoList.size(); i++){
			Node target = (Node)rmoList.get(i);
			target = getAncestorNode(rmbLength, target);

			if (ENABLE_DUPLICATE_DETECTION == true){
				if (target != lastocc){
					rootCount++;
					lastocc = target;
				}
			}
			else{
				if (nodeMemberReverse(target, rootOccList) == false) {
					rootCount++;
					rootOccList.add(target);
				}
				//Note: rootCount == rootOccList.size() must hold.
			}
		}
		return rootCount;
    }

    static Node getAncestorNode(int p, Node node){

		while (p > 0){
			if (node == null)
				return null;
			node = node.getParentNode();
			p--;
		}
		return node;
    }

    //**************************************************
    // Pruning by node and edge frequencies
    //**************************************************

    static void addForbiddenNodeSet(String nodename){

        ForbiddenNodeSet.add(new String(nodename));
    }

    static boolean isSafeNode(String nodename){

        if (ForbiddenNodeSet.contains(nodename))
            return false;
        else
            return true;
    }

    static void addForbiddenEdgeSet(String pname,String cname){

        if (ForbiddenEdgeSet.containsKey(pname) == false)
			ForbiddenEdgeSet.put(new String(pname),new HashSet());

		HashSet cutchild = (HashSet)ForbiddenEdgeSet.get(pname);
        cutchild.add(new String(cname));
    }

    static boolean isSafeEdge(String pname,String cname){

        if (ForbiddenEdgeSet.containsKey(pname)){
			HashSet cutchild = (HashSet)ForbiddenEdgeSet.get(pname);
            if (cutchild.contains(cname))
                return false;
            else
                return true;
        }
        else
            return true;
    }

    //**************************************************
    // Utilities
    //**************************************************

    static boolean nodeMemberReverse(Node target, NodeList nodelist){
		//Linear search in reverse order.
        int i;
        Node current;

        for (i = (nodelist.size()-1); i >= 0; i--){
            current = (Node)nodelist.get(i);
            if (current == target)
                return true;
        }
        return false;
    }

    static boolean nodeMemberForward(String nodename, NodeList nodelist){
		//For comparison
        for (int i = 0; i < nodelist.size(); i++){
            String name = (String)nodelist.get(i);
            if (name.equals(nodename))
                return true;
        }

        return false;
    }

    static String getXMLCommentString(String text) {
		return "<!-- "+text+" -->";
    }

    static String getXMLElementString(String name, String text) {
		return "<"+name+">"+text+"</"+name+">";
    }

    static String xmlAValInt(int x) {
		return "\""+String.valueOf(x)+"\"";
    }

    static int getNumChildren(Node parent) {
		int count = 0;
		if (parent == null)
			Util.errorExit("getNumChildren: null parent node!");
		Node target = parent.getFirstChild();
		while (target != null) {
			count++; 
			target = target.getNextSibling();
		}
		return count;
    }


    //**************************************************
    // Concept tree: Generalized labels with hierarchy
    //**************************************************

    public static void setCTPath(File file) throws Exception{

		int i;
		String s,key;
		String[] value;
		BufferedReader fin = new BufferedReader(new FileReader(file));
		while ((s=fin.readLine()) != null){
			i = 0;
			StringTokenizer st = new StringTokenizer(s," \t");
			if (st.countTokens() > 1){
				key = st.nextToken();
				value = new String[st.countTokens()];
				while (st.countTokens() > 0)
					value[i++] = st.nextToken();
				ctPathSet.put(key,value);
			}
		}
		fin.close();
    }
	
    public static String[] getCTPath(String tagname){

		if (ctPathSet.containsKey(tagname) == true)
			return (String[])ctPathSet.get(tagname);
		else
			return new String[]{};
    }

}

// EOF 

